defmodule Membrane.Core.Element.LifecycleController do
  @moduledoc false

  # Module handling element initialization, termination, playback state changes
  # and similar stuff.

  use Bunch
  use Membrane.Core.PlaybackHandler

  alias Membrane.{Clock, Element, Sync}
  alias Membrane.Core.{CallbackHandler, Message}
  alias Membrane.Core.Child.PadController
  alias Membrane.Core.Element.{ActionHandler, PlaybackBuffer, State}
  alias Membrane.Element.CallbackContext

  require Membrane.Core.Child.PadModel
  require Membrane.Core.Message
  require Membrane.Core.Playback
  require Membrane.Logger

  @doc """
  Performs initialization tasks and executes `handle_init` callback.
  """
  @spec handle_init(Element.options_t(), State.t()) :: State.stateful_try_t()
  def handle_init(options, %State{module: module} = state) do
    Membrane.Logger.debug(
      "Initializing element: #{inspect(module)}, options: #{inspect(options)}"
    )

    :ok = Sync.register(state.synchronization.stream_sync)

    state =
      if Bunch.Module.check_behaviour(module, :membrane_clock?) do
        {:ok, clock} = Clock.start_link()
        put_in(state.synchronization.clock, clock)
      else
        state
      end

    with {:ok, state} <-
           CallbackHandler.exec_and_handle_callback(
             :handle_init,
             ActionHandler,
             %{state: false},
             [options],
             state
           ) do
      Membrane.Logger.debug("Element initialized: #{inspect(module)}")
      {:ok, state}
    else
      {{:error, reason}, state} ->
        Membrane.Logger.error("""
        Failed to initialize element
        Reason: #{inspect(reason)}
        State: #{inspect(state, pretty: true)}
        """)

        {{:error, reason}, state}
    end
  end

  @doc """
  Performs shutdown checks and executes `handle_shutdown` callback.
  """
  @spec handle_shutdown(reason :: any, State.t()) :: {:ok, State.t()}
  def handle_shutdown(reason, state) do
    playback_state = state.playback.state

    if playback_state == :terminating do
      Membrane.Logger.debug("Terminating element, reason: #{inspect(reason)}")
    else
      Membrane.Logger.warn("""
      Terminating element possibly not prepared for termination as it was in state #{
        inspect(playback_state)
      }.
      Reason: #{inspect(reason)}",
      State: #{inspect(state, pretty: true)}
      """)
    end

    %State{module: module, internal_state: internal_state} = state
    :ok = module.handle_shutdown(reason, internal_state)
    {:ok, state}
  end

  @spec handle_pipeline_down(reason :: any, State.t()) :: {:ok, State.t()}
  def handle_pipeline_down(reason, state) do
    if reason != :normal do
      Membrane.Logger.error("""
      Shutting down because of pipeline failure
      Reason: #{inspect(reason)}
      State: #{inspect(state, pretty: true)}
      """)
    end

    handle_shutdown(reason, state)
  end

  @doc """
  Handles custom messages incoming to element.
  """
  @spec handle_other(message :: any, State.t()) :: State.stateful_try_t()
  def handle_other(message, state) do
    require CallbackContext.Other
    context = &CallbackContext.Other.from_state/1

    CallbackHandler.exec_and_handle_callback(
      :handle_other,
      ActionHandler,
      %{context: context},
      [message],
      state
    )
  end

  @impl PlaybackHandler
  def handle_playback_state(old_playback_state, new_playback_state, state) do
    require CallbackContext.PlaybackChange
    context = &CallbackContext.PlaybackChange.from_state/1
    callback = PlaybackHandler.state_change_callback(old_playback_state, new_playback_state)

    state =
      if old_playback_state == :playing and new_playback_state == :prepared do
        state.pads.data
        |> Map.values()
        |> Enum.filter(&(&1.direction == :input))
        |> Enum.map(& &1.ref)
        |> Enum.reduce(state, fn pad, state_acc ->
          {:ok, state} = PadController.generate_eos_if_needed(pad, state_acc)
          state
        end)
      else
        state
      end

    if callback do
      CallbackHandler.exec_and_handle_callback(
        callback,
        ActionHandler,
        %{context: context},
        [],
        state
      )
    else
      {:ok, state}
    end
  end

  @impl PlaybackHandler
  def handle_playback_state_changed(_old, new, state) do
    if new == :stopped, do: unlink(state.pads.data)

    with {:ok, state} <- PlaybackBuffer.eval(state) do
      if new == :terminating,
        do: {:stop, :normal, state},
        else: {:ok, state}
    end
  end

  defp unlink(pads_data) do
    pads_data
    |> Map.values()
    |> Enum.each(&Message.send(&1.pid, :handle_unlink, &1.other_ref))
  end
end
